import keras
import numpy as np
from os.path import join, isfile
import scipy.misc
from os import listdir
import tifffile

def process_X(X):
    X = X.astype(float)
    X -= 128
    X /= 128

    if (len(X.shape) == 3):
        X = X[:, :, :, np.newaxis]

    return X

def process_Xy(X, y):
    X = process_X(X)
    if (len(y.shape) == 1):
        y = y[:, np.newaxis]
    #if (len(y.shape) == 3):
    #    y = y[:, :, :, np.newaxis]
    return X, y

def get_config(config, key, default=None):
    if (hasattr(config, key)):
        return getattr(config, key)
    return default
class DataGenerator(keras.utils.Sequence):
    path = None#'patches_fast/'



    'Generates data for Keras'
    def __init__(self, path, list_IDs, config, folder):
        'Initialization'
        self.path = path
        self.batch_size = get_config(config, 'batch_size', 32)
        self.list_IDs = list_IDs
        self.shuffle = get_config(config, 'shuffle', True)
        self.process_patch = get_config(config, 'process_patch', None)
        self.max_items_per_epoch = get_config(config, 'max_items_per_epoch', None)
        self.process_patch_full = get_config(config, 'process_patch_full', None)
        self.on_epoch_end_cb = get_config(config, 'on_epoch_end_cb', None)
        self.folder = folder
        self.image_extension = get_config(config, 'image_extension', '.png')
        self.on_epoch_end()

    def __len__(self):
        'Denotes the number of batches per epoch'
        return int(np.floor(len(self.indexes) / self.batch_size))

    def __getitem__(self, index):
        'Generate one batch of data'
        # Generate indexes of the batch
        
        list_IDs_temp = self.indexes[index*self.batch_size:(index+1)*self.batch_size]

        # Find list of IDs
        #list_IDs_temp = [self.list_IDs[k] for k in indexes]

        # Generate data
        X, comp_X, y = self.__data_generation(list_IDs_temp)

        if (comp_X.shape[0] == 0 or comp_X[0] is None):
            return X, y
        else:
            return [X, comp_X], y

    def on_epoch_end(self):
        'Updates indexes after each epoch'
        if (self.on_epoch_end_cb is None):
            self.indexes = self.list_IDs.copy() #np.arange(len(self.list_IDs))
            if self.shuffle == True:
                np.random.shuffle(self.indexes)
            if (self.max_items_per_epoch is not None and len(self.indexes) > self.max_items_per_epoch):
                self.indexes = self.indexes[:self.max_items_per_epoch]
        else:
            self.on_epoch_end_cb(self)


    def __data_generation(self, list_IDs_temp, randomize=True):
        path = self.path
        'Generates data containing batch_size samples'
        # Initialization
        X = []
        y = []
        comp_X = []
        
        # Generate data
        for i, ID in enumerate(list_IDs_temp):
            if (self.image_extension == '.tif'):
                im_np = tifffile.imread(join(path, str(ID) + '_X' + self.image_extension))
            elif (self.image_extension == '.npz'):
                im_np = np.load(join(path, str(ID) + '_X' + self.image_extension))['X']
            else:
                im_np = scipy.misc.imread(join(path, str(ID) + '_X' + self.image_extension))
            if (self.process_patch is not None):
                im_np = self.process_patch(im_np, True)
                
            y_np = None
            y_filepath = join(path, str(ID) + '_y.npz')
            if (isfile(y_filepath)):
                y_np = np.load(y_filepath)['y'] #.astype(float)
            
            if (self.process_patch_full is not None):
                im_np, y_np = self.process_patch_full(im_np, y_np, ID)

            comp_X_filepath = join(path, str(ID) + '_comp_X.npz')
            comp_X_np = None
            if (isfile(comp_X_filepath)):
                comp_X_np = np.load(comp_X_filepath)['complementary_X'].tolist()
                comp_X_np = np.hstack(list(comp_X_np.values()))
            X.append(im_np)
            comp_X.append(comp_X_np)
            y.append(y_np)
        X = np.array(X)
        comp_X = np.array(comp_X)
        y = np.array(y)
        X, y = process_Xy(X, y)
        return X, comp_X, y

    @staticmethod
    def get_ids(path, config):
        image_extension = get_config(config, 'image_extension', '.png')
        files = [f for f in listdir(path) if isfile(join(path, f))]
        files = sorted(files)
        ids = []
        for file in files:
            if (file.endswith('_X'+image_extension)):
                ids.append(file[:-6])
                
        init_ids = list(ids)
        return np.array(ids)